@@ -2,6 +2,7 @@ import asyncio
 import codecs
 import io
 import json
+import re
 import sys
 import traceback
 import warnings
@@ -22,7 +23,6 @@ from .formdata import FormData
 from .helpers import PY_36, HeadersMixin, TimerNoop, noop, reify, set_result
 from .http import SERVER_SOFTWARE, HttpVersion10, HttpVersion11, StreamWriter
 from .log import client_logger
-from .streams import StreamReader
 
 
 try:
@@ -39,6 +39,9 @@ except ImportError:  # pragma: no cover
 __all__ = ('ClientRequest', 'ClientResponse', 'RequestInfo', 'Fingerprint')
 
 
+json_re = re.compile('^application/(?:[\w.+-]+?\+)?json')
+
+
 @attr.s(frozen=True, slots=True)
 class ContentDisposition:
     type = attr.ib(type=str)
@@ -51,6 +54,11 @@ class RequestInfo:
     url = attr.ib(type=URL)
     method = attr.ib(type=str)
     headers = attr.ib(type=CIMultiDictProxy)
+    real_url = attr.ib(type=URL)
+
+    @real_url.default
+    def real_url_default(self):
+        return self.url
 
 
 class Fingerprint:
@@ -131,6 +139,12 @@ def _merge_ssl_params(ssl, verify_ssl, ssl_context, fingerprint):
 ConnectionKey = namedtuple('ConnectionKey', ['host', 'port', 'ssl'])
 
 
+def _is_expected_content_type(response_content_type, expected_content_type):
+    if expected_content_type == 'application/json':
+        return json_re.match(response_content_type)
+    return expected_content_type in response_content_type
+
+
 class ClientRequest:
     GET_METHODS = {
         hdrs.METH_GET,
@@ -168,7 +182,8 @@ class ClientRequest:
                  proxy=None, proxy_auth=None,
                  timer=None, session=None, auto_decompress=True,
                  ssl=None,
-                 proxy_headers=None):
+                 proxy_headers=None,
+                 traces=None):
 
         if loop is None:
             loop = asyncio.get_event_loop()
@@ -181,8 +196,8 @@ class ClientRequest:
             url2 = url.with_query(params)
             q.extend(url2.query)
             url = url.with_query(q)
-        self.url = url.with_fragment(None)
         self.original_url = url
+        self.url = url.with_fragment(None)
         self.method = method.upper()
         self.chunked = chunked
         self.compress = compress
@@ -209,6 +224,9 @@ class ClientRequest:
         if data or self.method not in self.GET_METHODS:
             self.update_transfer_encoding()
         self.update_expect_continue(expect100)
+        if traces is None:
+            traces = []
+        self._traces = traces
 
     def is_ssl(self):
         return self.url.scheme in ('https', 'wss')
@@ -231,7 +249,8 @@ class ClientRequest:
 
     @property
     def request_info(self):
-        return RequestInfo(self.url, self.method, self.headers)
+        return RequestInfo(self.url, self.method,
+                           self.headers, self.original_url)
 
     def update_host(self, url):
         """Update destination host, port and connection type (ssl)."""
@@ -443,7 +462,7 @@ class ClientRequest:
                     self.body = (self.body,)
 
                 for chunk in self.body:
-                    writer.write(chunk)
+                    await writer.write(chunk)
 
             await writer.write_eof()
         except OSError as exc:
@@ -461,7 +480,7 @@ class ClientRequest:
         finally:
             self._writer = None
 
-    def send(self, conn):
+    async def send(self, conn):
         # Specify request target:
         # - CONNECT request must send authority form URI
         # - not CONNECT proxy must send absolute form URI
@@ -475,7 +494,10 @@ class ClientRequest:
             if self.url.raw_query_string:
                 path += '?' + self.url.raw_query_string
 
-        writer = StreamWriter(conn.protocol, conn.transport, self.loop)
+        writer = StreamWriter(
+            conn.protocol, self.loop,
+            on_chunk_sent=self._on_chunk_request_sent
+        )
 
         if self.compress:
             writer.enable_compression(self.compress)
@@ -503,9 +525,9 @@ class ClientRequest:
             self.headers[hdrs.CONNECTION] = connection
 
         # status + headers
-        status_line = '{0} {1} HTTP/{2[0]}.{2[1]}\r\n'.format(
+        status_line = '{0} {1} HTTP/{2[0]}.{2[1]}'.format(
             self.method, path, self.version)
-        writer.write_headers(status_line, self.headers)
+        await writer.write_headers(status_line, self.headers)
 
         self._writer = self.loop.create_task(self.write_bytes(writer, conn))
 
@@ -513,9 +535,11 @@ class ClientRequest:
             self.method, self.original_url,
             writer=self._writer, continue100=self._continue, timer=self._timer,
             request_info=self.request_info,
-            auto_decompress=self._auto_decompress)
-
-        self.response._post_init(self.loop, self._session)
+            auto_decompress=self._auto_decompress,
+            traces=self._traces,
+            loop=self.loop,
+            session=self._session
+        )
         return self.response
 
     async def close(self):
@@ -531,6 +555,10 @@ class ClientRequest:
                 self._writer.cancel()
             self._writer = None
 
+    async def _on_chunk_request_sent(self, chunk):
+        for trace in self._traces:
+            await trace.send_request_chunk_sent(chunk)
+
 
 class ClientResponse(HeadersMixin):
 
@@ -544,34 +572,38 @@ class ClientResponse(HeadersMixin):
     raw_headers = None  # Response raw headers, a sequence of pairs
 
     _connection = None  # current connection
-    flow_control_class = StreamReader  # reader flow control
     _reader = None     # input stream
     _source_traceback = None
     # setted up by ClientRequest after ClientResponse object creation
     # post-init stage allows to not change ctor signature
-    _loop = None
     _closed = True  # to allow __del__ for non-initialized properly response
-    _session = None
 
     def __init__(self, method, url, *,
-                 writer=None, continue100=None, timer=None,
-                 request_info=None, auto_decompress=True):
+                 writer, continue100, timer,
+                 request_info, auto_decompress,
+                 traces, loop, session):
         assert isinstance(url, URL)
 
         self.method = method
         self.headers = None
         self.cookies = SimpleCookie()
 
-        self._url = url
-        self._content = None
+        self._real_url = url
+        self._url = url.with_fragment(None)
+        self._body = None
         self._writer = writer
-        self._continue = continue100
+        self._continue = continue100  # None by default
         self._closed = True
         self._history = ()
         self._request_info = request_info
         self._timer = timer if timer is not None else TimerNoop()
-        self._auto_decompress = auto_decompress
-        self._cache = {}  # reqired for @reify method decorator
+        self._auto_decompress = auto_decompress  # True by default
+        self._cache = {}  # required for @reify method decorator
+        self._traces = traces
+        self._loop = loop
+        self._session = session  # store a reference to session #1985
+        if loop.get_debug():
+            self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
     @property
     def url(self):
@@ -583,6 +615,10 @@ class ClientResponse(HeadersMixin):
             "Deprecated, use .url #1654", DeprecationWarning, stacklevel=2)
         return self._url
 
+    @property
+    def real_url(self):
+        return self._real_url
+
     @property
     def host(self):
         return self._url.host
@@ -605,15 +641,7 @@ class ClientResponse(HeadersMixin):
         filename = multipart.content_disposition_filename(params)
         return ContentDisposition(disposition_type, params, filename)
 
-    def _post_init(self, loop, session):
-        self._loop = loop
-        self._session = session  # store a reference to session #1985
-        if loop.get_debug():
-            self._source_traceback = traceback.extract_stack(sys._getframe(1))
-
     def __del__(self, _warnings=warnings):
-        if self._loop is None:
-            return  # not started
         if self._closed:
             return
 
@@ -659,6 +687,37 @@ class ClientResponse(HeadersMixin):
         """A sequence of of responses, if redirects occurred."""
         return self._history
 
+    @property
+    def links(self):
+        links_str = ", ".join(self.headers.getall("link", []))
+
+        links = MultiDict()
+
+        if not links_str:
+            return MultiDictProxy(links)
+
+        for val in re.split(r",(?=\s*<)", links_str):
+            url, params = re.match(r"\s*<(.*)>(.*)", val).groups()
+            params = params.split(";")[1:]
+
+            link = MultiDict()
+
+            for param in params:
+                key, _, value, _ = re.match(
+                    r"^\s*(\S*)\s*=\s*(['\"]?)(.*?)(\2)\s*$",
+                    param, re.M
+                ).groups()
+
+                link.add(key, value)
+
+            key = link.get("rel", url)
+
+            link.add("url", self.url.join(URL(url)))
+
+            links.add(key, MultiDictProxy(link))
+
+        return MultiDictProxy(links)
+
     async def start(self, connection, read_until_eof=False):
         """Start response processing."""
         self._closed = False
@@ -679,7 +738,7 @@ class ClientResponse(HeadersMixin):
                 except http.HttpProcessingError as exc:
                     raise ClientResponseError(
                         self.request_info, self.history,
-                        code=exc.code,
+                        status=exc.code,
                         message=exc.message, headers=exc.headers) from exc
 
                 if (message.code < 100 or
@@ -767,7 +826,7 @@ class ClientResponse(HeadersMixin):
             raise ClientResponseError(
                 self.request_info,
                 self.history,
-                code=self.status,
+                status=self.status,
                 message=self.reason,
                 headers=self.headers)
 
@@ -793,14 +852,16 @@ class ClientResponse(HeadersMixin):
 
     async def read(self):
         """Read response payload."""
-        if self._content is None:
+        if self._body is None:
             try:
-                self._content = await self.content.read()
+                self._body = await self.content.read()
+                for trace in self._traces:
+                    await trace.send_response_chunk_received(self._body)
             except BaseException:
                 self.close()
                 raise
 
-        return self._content
+        return self._body
 
     def get_encoding(self):
         ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
@@ -817,7 +878,7 @@ class ClientResponse(HeadersMixin):
                 # RFC 7159 states that the default encoding is UTF-8.
                 encoding = 'utf-8'
             else:
-                encoding = chardet.detect(self._content)['encoding']
+                encoding = chardet.detect(self._body)['encoding']
         if not encoding:
             encoding = 'utf-8'
 
@@ -825,23 +886,23 @@ class ClientResponse(HeadersMixin):
 
     async def text(self, encoding=None, errors='strict'):
         """Read response payload and decode."""
-        if self._content is None:
+        if self._body is None:
             await self.read()
 
         if encoding is None:
             encoding = self.get_encoding()
 
-        return self._content.decode(encoding, errors=errors)
+        return self._body.decode(encoding, errors=errors)
 
     async def json(self, *, encoding=None, loads=json.loads,
                    content_type='application/json'):
         """Read and decodes JSON response."""
-        if self._content is None:
+        if self._body is None:
             await self.read()
 
         if content_type:
             ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
-            if content_type not in ctype:
+            if not _is_expected_content_type(ctype, content_type):
                 raise ContentTypeError(
                     self.request_info,
                     self.history,
@@ -849,7 +910,7 @@ class ClientResponse(HeadersMixin):
                              'unexpected mimetype: %s' % ctype),
                     headers=self.headers)
 
-        stripped = self._content.strip()
+        stripped = self._body.strip()
         if not stripped:
             return None
 
