@@ -6,6 +6,7 @@ import sys
 from unittest import mock
 
 import pytest
+from multidict import CIMultiDict
 from yarl import URL
 
 import aiohttp
@@ -1018,3 +1019,174 @@ async def test_response_read_triggers_callback(loop, session):
         trace.send_response_chunk_received.call_args ==
         mock.call(response_body)
     )
+
+
+def test_response_real_url(loop, session):
+    url = URL('http://def-cl-resp.org/#urlfragment')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    assert response.url == url.with_fragment(None)
+    assert response.real_url == url
+
+
+def test_response_links_comma_separated(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict([
+        (
+            "Link",
+            ('<http://example.com/page/1.html>; rel=next, '
+             '<http://example.com/>; rel=home')
+        )
+    ])
+    assert (
+        response.links ==
+        {'next':
+         {'url': URL('http://example.com/page/1.html'),
+          'rel': 'next'},
+         'home':
+         {'url': URL('http://example.com/'),
+          'rel': 'home'}
+         }
+    )
+
+
+def test_response_links_multiple_headers(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict([
+        (
+            "Link",
+            '<http://example.com/page/1.html>; rel=next'
+        ),
+        (
+            "Link",
+            '<http://example.com/>; rel=home'
+        )
+    ])
+    assert (
+        response.links ==
+        {'next':
+         {'url': URL('http://example.com/page/1.html'),
+          'rel': 'next'},
+         'home':
+         {'url': URL('http://example.com/'),
+          'rel': 'home'}
+         }
+    )
+
+
+def test_response_links_no_rel(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict([
+        (
+            "Link",
+            '<http://example.com/>'
+        )
+    ])
+    assert (
+        response.links ==
+        {
+            'http://example.com/':
+            {'url': URL('http://example.com/')}
+        }
+    )
+
+
+def test_response_links_quoted(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict([
+        (
+            "Link",
+            '<http://example.com/>; rel="home-page"'
+        ),
+    ])
+    assert (
+        response.links ==
+        {'home-page':
+         {'url': URL('http://example.com/'),
+          'rel': 'home-page'}
+         }
+    )
+
+
+def test_response_links_relative(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict([
+        (
+            "Link",
+            '</relative/path>; rel=rel'
+        ),
+    ])
+    assert (
+        response.links ==
+        {'rel':
+         {'url': URL('http://def-cl-resp.org/relative/path'),
+          'rel': 'rel'}
+         }
+    )
+
+
+def test_response_links_empty(loop, session):
+    url = URL('http://def-cl-resp.org/')
+    response = ClientResponse('get', url,
+                              request_info=mock.Mock(),
+                              writer=mock.Mock(),
+                              continue100=None,
+                              timer=TimerNoop(),
+                              auto_decompress=True,
+                              traces=[],
+                              loop=loop,
+                              session=session)
+    response.headers = CIMultiDict()
+    assert response.links == {}
