@@ -8,6 +8,7 @@ import platform
 import socket
 import ssl
 import uuid
+from collections import deque
 from unittest import mock
 
 import pytest
@@ -40,8 +41,8 @@ def ssl_key():
 
 
 @pytest.fixture
-def unix_sockname(tmpdir):
-    sock_path = tmpdir / 'socket.sock'
+def unix_sockname(shorttmpdir):
+    sock_path = shorttmpdir / 'socket.sock'
     return str(sock_path)
 
 
@@ -62,6 +63,81 @@ def unix_server(loop, unix_sockname):
         loop.run_until_complete(runner.cleanup())
 
 
+def test_connection_del(loop):
+    connector = mock.Mock()
+    key = mock.Mock()
+    protocol = mock.Mock()
+    loop.set_debug(0)
+    conn = Connection(connector, key, protocol, loop=loop)
+    exc_handler = mock.Mock()
+    loop.set_exception_handler(exc_handler)
+
+    with pytest.warns(ResourceWarning):
+        del conn
+        gc.collect()
+
+    connector._release.assert_called_with(
+        key,
+        protocol,
+        should_close=True
+    )
+    msg = {
+        'message': mock.ANY,
+        'client_connection': mock.ANY,
+    }
+    exc_handler.assert_called_with(loop, msg)
+
+
+def test_connection_del_loop_debug(loop):
+    connector = mock.Mock()
+    key = mock.Mock()
+    protocol = mock.Mock()
+    loop.set_debug(1)
+    conn = Connection(connector, key, protocol, loop=loop)
+    exc_handler = mock.Mock()
+    loop.set_exception_handler(exc_handler)
+
+    with pytest.warns(ResourceWarning):
+        del conn
+        gc.collect()
+
+    msg = {
+        'message': mock.ANY,
+        'client_connection': mock.ANY,
+        'source_traceback': mock.ANY
+    }
+    exc_handler.assert_called_with(loop, msg)
+
+
+def test_connection_del_loop_closed(loop):
+    connector = mock.Mock()
+    key = mock.Mock()
+    protocol = mock.Mock()
+    loop.set_debug(1)
+    conn = Connection(connector, key, protocol, loop=loop)
+    exc_handler = mock.Mock()
+    loop.set_exception_handler(exc_handler)
+    loop.close()
+
+    with pytest.warns(ResourceWarning):
+        del conn
+        gc.collect()
+
+    assert not connector._release.called
+    assert not exc_handler.called
+
+
+def test_connection_detach(loop):
+    connector = mock.Mock()
+    key = mock.Mock()
+    protocol = mock.Mock()
+    conn = Connection(connector, key, protocol, loop=loop)
+    conn._notify_release = mock.Mock()
+    conn.detach()
+    assert conn._notify_release.called
+    connector._release_acquired.assert_called_with(protocol)
+
+
 def test_del(loop):
     conn = aiohttp.BaseConnector(loop=loop)
     proto = mock.Mock(should_close=False)
@@ -90,7 +166,7 @@ async def test_del_with_scheduled_cleanup(loop):
     loop.set_debug(True)
     conn = aiohttp.BaseConnector(loop=loop, keepalive_timeout=0.01)
     transp = mock.Mock()
-    conn._conns['a'] = [(transp, 'proto', 123)]
+    conn._conns['a'] = [(transp, 123)]
 
     conns_impl = conn._conns
     exc_handler = mock.Mock()
@@ -115,7 +191,7 @@ async def test_del_with_scheduled_cleanup(loop):
 def test_del_with_closed_loop(loop):
     conn = aiohttp.BaseConnector(loop=loop)
     transp = mock.Mock()
-    conn._conns['a'] = [(transp, 'proto', 123)]
+    conn._conns['a'] = [(transp, 123)]
 
     conns_impl = conn._conns
     exc_handler = mock.Mock()
@@ -291,18 +367,19 @@ def test_release_already_closed(loop):
     assert not conn._release_acquired.called
 
 
-def test_release_waiter(loop, key, key2):
+def test_release_waiter_no_limit(loop, key, key2):
     # limit is 0
     conn = aiohttp.BaseConnector(limit=0, loop=loop)
     w = mock.Mock()
     w.done.return_value = False
     conn._waiters[key].append(w)
     conn._release_waiter()
-    assert len(conn._waiters) == 1
-    assert not w.done.called
+    assert len(conn._waiters[key]) == 0
+    assert w.done.called
     conn.close()
 
-    # release first available
+
+def test_release_waiter_first_available(loop, key, key2):
     conn = aiohttp.BaseConnector(loop=loop)
     w1, w2 = mock.Mock(), mock.Mock()
     w1.done.return_value = False
@@ -314,26 +391,28 @@ def test_release_waiter(loop, key, key2):
             not w1.set_result.called and w2.set_result.called)
     conn.close()
 
-    # limited available
+
+def test_release_waiter_release_first(loop, key, key2):
     conn = aiohttp.BaseConnector(loop=loop, limit=1)
     w1, w2 = mock.Mock(), mock.Mock()
     w1.done.return_value = False
     w2.done.return_value = False
-    conn._waiters[key] = [w1, w2]
+    conn._waiters[key] = deque([w1, w2])
     conn._release_waiter()
     assert w1.set_result.called
     assert not w2.set_result.called
     conn.close()
 
-    # limited available
+
+def test_release_waiter_skip_done_waiter(loop, key, key2):
     conn = aiohttp.BaseConnector(loop=loop, limit=1)
     w1, w2 = mock.Mock(), mock.Mock()
     w1.done.return_value = True
     w2.done.return_value = False
-    conn._waiters[key] = [w1, w2]
+    conn._waiters[key] = deque([w1, w2])
     conn._release_waiter()
     assert not w1.set_result.called
-    assert not w2.set_result.called
+    assert w2.set_result.called
     conn.close()
 
 
@@ -343,14 +422,27 @@ def test_release_waiter_per_host(loop, key, key2):
     w1, w2 = mock.Mock(), mock.Mock()
     w1.done.return_value = False
     w2.done.return_value = False
-    conn._waiters[key] = [w1]
-    conn._waiters[key2] = [w2]
+    conn._waiters[key] = deque([w1])
+    conn._waiters[key2] = deque([w2])
     conn._release_waiter()
     assert ((w1.set_result.called and not w2.set_result.called) or
             (not w1.set_result.called and w2.set_result.called))
     conn.close()
 
 
+def test_release_waiter_no_available(loop, key, key2):
+    # limit is 0
+    conn = aiohttp.BaseConnector(limit=0, loop=loop)
+    w = mock.Mock()
+    w.done.return_value = False
+    conn._waiters[key].append(w)
+    conn._available_connections = mock.Mock(return_value=0)
+    conn._release_waiter()
+    assert len(conn._waiters) == 1
+    assert not w.done.called
+    conn.close()
+
+
 def test_release_close(loop):
     conn = aiohttp.BaseConnector(loop=loop)
     proto = mock.Mock(should_close=True)
@@ -960,7 +1052,9 @@ async def test_connect_tracing(loop):
     conn._create_connection.return_value = loop.create_future()
     conn._create_connection.return_value.set_result(proto)
 
-    await conn.connect(req, traces=traces)
+    conn2 = await conn.connect(req, traces=traces)
+    conn2.release()
+
     on_connection_create_start.assert_called_with(
         session,
         trace_config_ctx,
@@ -1411,7 +1505,8 @@ async def test_connect_reuseconn_tracing(loop, key):
 
     conn = aiohttp.BaseConnector(loop=loop, limit=1)
     conn._conns[key] = [(proto, loop.time())]
-    await conn.connect(req, traces=traces)
+    conn2 = await conn.connect(req, traces=traces)
+    conn2.release()
 
     on_connection_reuseconn.assert_called_with(
         session,
@@ -1626,6 +1721,48 @@ async def test_connect_with_limit_concurrent(loop):
     assert max_connections == num_connections
 
 
+async def test_connect_waiters_cleanup(loop):
+    proto = mock.Mock()
+    proto.is_connected.return_value = True
+
+    req = ClientRequest('GET', URL('http://host:80'), loop=loop)
+
+    conn = aiohttp.BaseConnector(loop=loop, limit=1)
+    conn._available_connections = mock.Mock(return_value=0)
+
+    t = loop.create_task(conn.connect(req))
+
+    await asyncio.sleep(0, loop=loop)
+    assert conn._waiters.keys()
+
+    t.cancel()
+    await asyncio.sleep(0, loop=loop)
+    assert not conn._waiters.keys()
+
+
+async def test_connect_waiters_cleanup_key_error(loop):
+    proto = mock.Mock()
+    proto.is_connected.return_value = True
+
+    req = ClientRequest('GET', URL('http://host:80'), loop=loop)
+
+    conn = aiohttp.BaseConnector(loop=loop, limit=1)
+    conn._available_connections = mock.Mock(return_value=0)
+
+    t = loop.create_task(conn.connect(req))
+
+    await asyncio.sleep(0, loop=loop)
+    assert conn._waiters.keys()
+
+    # we delete the entry explicitly before the
+    # canceled connection grabs the loop again, we
+    # must expect a none failure termination
+    conn._waiters.clear()
+    t.cancel()
+    await asyncio.sleep(0, loop=loop)
+    assert not conn._waiters.keys() == []
+
+
 async def test_close_with_acquired_connection(loop):
     proto = mock.Mock()
     proto.is_connected.return_value = True
@@ -1736,6 +1873,29 @@ async def test_error_on_connection(loop):
     assert ret._key == 'key'
     assert ret.protocol == proto
     assert proto in conn._acquired
+    ret.release()
+
+
+async def test_cancelled_waiter(loop):
+    conn = aiohttp.BaseConnector(limit=1, loop=loop)
+    req = mock.Mock()
+    req.connection_key = 'key'
+    proto = mock.Mock()
+
+    async def create_connection(req, traces=None):
+        await asyncio.sleep(1)
+        return proto
+
+    conn._create_connection = create_connection
+
+    conn._acquired.add(proto)
+
+    conn2 = loop.create_task(conn.connect(req))
+    await asyncio.sleep(0, loop=loop)
+    conn2.cancel()
+
+    with pytest.raises(asyncio.CancelledError):
+        await conn2
 
 
 async def test_error_on_connection_with_cancelled_waiter(loop):
@@ -1785,6 +1945,7 @@ async def test_error_on_connection_with_cancelled_waiter(loop):
     assert ret._key == 'key'
     assert ret.protocol == proto
     assert proto in conn._acquired
+    ret.release()
 
 
 async def test_tcp_connector(aiohttp_client, loop):
